.editor {
    height: 400px;
    width: 49.7%;
    display: inline-block;
}
.configurator {
    height: 200px;
    width: 100%;
    clear: both;
    float: left;
}
#liveTester {
    background: black;
    color: white;
}
#sideBySide {
    white-space: nowrap;
}
#liveTester.fullscreen {
    position: absolute;
    top: 0px;
    left: 0px;
    right: 0px;
    z-index: 99;
    bottom: 0px;
}
